#!/usr/bin/env python

import vtk
import time
import random
from numpy import *
from scipy.io import read_array

## Read the basic configuration of the worm.
## Later, its node coordinates are modified
## with a vtkProgrammableFilter
reader = vtk.vtkUnstructuredGridReader()
reader.SetFileName("worm.vtk")

## The filter to modify the coordinates of the worm
change_coords_filter = vtk.vtkProgrammableFilter()
change_coords_filter.SetInputConnection(reader.GetOutputPort())

## Coordinate system: elevation is `y' axis. The worm displaces
## in the `z' direction. 

## A number of cones are added to create a landscape where the worm
## moves. The following parameters are defined.
##
## The size of the landscape (the part of the domain where cones are
## defined).
Lscape = 150
## Number of cones
ncones = 300
## Radius of cones bases are generated randomly in
## this range
Rmaxcone = 5.0
Rmincone = 1
## Radius of the worm 
Rworm = 1
## Cones are not added in |x|<2*Dmin
Dmin = Rworm + Rmaxcone

## This routine makes a cone of height H and
## basis radius R positioned at [x[0],0,x[1]]
## The routine returns the cone actor. 
def make_cone(x,H,R):
    ## Creates the cone
    cone = vtk.vtkConeSource()
    cone.SetHeight(H)
    cone.SetRadius(R)
    cone.SetResolution(20)
    cone.SetDirection(0,1,0)

    ## Creates the mapper
    coneMapper = vtk.vtkPolyDataMapper()
    coneMapper.SetInputConnection(cone.GetOutputPort() )
    
    ## Creates the actor
    cone_actor = vtk.vtkActor()
    cone_actor.SetMapper( coneMapper )
    xx = [x[0],0.5*H,x[1]]
    cone_actor.SetPosition(xx)
    return cone_actor

## All the cones are gathered in an `assembly'
assembly = vtk.vtkAssembly()
## Also the are gathered in a Python array
## I think that this may be avoided we know how
## to iterate on the assembly. 
cones = []
for k in range(ncones):
    ## Generate a ramdom point in |x|,|z|<Lscape
    x = Lscape*(2.0*random.random()-1.0)
    ## Exclude a strip along the path of the worm (the z axis)
    if abs(x)<Dmin:
        x = Dmin*sign(x)
    z = Lscape*(2.0*random.random()-1.0)
    ## Generate random value of cone radius
    R = Rmincone+(Rmaxcone-Rmincone)*random.random()
    ## Aspect ratio of cones is fixed
    H = 3*R
    ## Make the cone
    cone = make_cone([x,z],H,R)
    ## Add cone to assembly and to `cones' Python array
    assembly.AddPart(cone)
    cones.append(cone)

## The algorithm of movement of the worm is as follows.  The state of
## the worm is defined by the positions of the extremes z1,z2.  At
## each time one of its extremmities are moving with velocity `speed'.
## Which of the extremmities is moving is stored in variable `moving'.
## When the length of the worm passes a threshold `Lmax' then we
## switch to moving=1 (the tail), and conversely if `L<Lmin' then we
## switch to moving the head (moving=2)

## Initial position of worm extremmities
z1 = 0.0
z2 = 9.0
## Min/Max length of worm
Lmin = 4.0
Lmax = 8.0
## Advance speed of one extremmity at a time
speed = 3.0
## Which extremmity is advancing now
moving = 1

## Time step
Dt = 0.2
## Initial time
t = 0.0
## Average speed of worm
mspeed = speed/2

## Position of worm node coordinates at undisturbed position
## (i.e. a plain cylinder)
X0 = read_array(file("worm.nod.tmp"))
## Displacements for the maximum distorted
## position of the worm, i.e. the position
## of the coordinates of the worm are X = X0+alpha*DX
## with 0<=alpha<=1
DX = read_array(file("worm.dx-nod.tmp"))

## This function is the kernel of the `vtkProgrammableFilter'
## It changes the coordinates of the worm points
def change_coords_fun():
    input = change_coords_filter.GetUnstructuredGridInput()
    numPts = input.GetNumberOfPoints()
    newPts = vtk.vtkPoints()

    ## Current length
    L = z2-z1
    ## Compute height such that 4*H**2+L**2 = Lmax**2
    ## This is some heuristic relation as if the length of the worm
    ## were fixed. 
    aux = (Lmax**2-L**2)/2
    if aux<0:
        aux = 0.0
    H = 0.25*sqrt(aux)

    ## Change the coordinates point by point
    for i in range(0, numPts):
        x = input.GetPoint(i)
        xx,yy,zz = X0[i] + H*DX[i]
        zz = z1+(L/Lmax)*zz
        newPts.InsertPoint(i, xx, yy, zz)
    output = change_coords_filter.GetUnstructuredGridOutput()
    output.CopyStructure(input)
    output.SetPoints(newPts)

change_coords_filter.SetExecuteMethod(change_coords_fun) 

# Create mapper
ugmapper = vtk.vtkDataSetMapper()
ugmapper.SetInputConnection(change_coords_filter.GetOutputPort())
ugmapper.ScalarVisibilityOff()

# Create actor
ugactor = vtk.vtkActor()
ugactor.SetMapper(ugmapper)
ugactor.GetProperty().SetDiffuseColor(1, 1, 1)

# Create the usual rendering stuff.
ren = vtk.vtkRenderer()
renWin = vtk.vtkRenderWindow()
renWin.AddRenderer(ren)
renWin.SetSize(640,480)

ren.SetBackground(.1, .2, .4)
ren.AddActor(ugactor)
ren.AddActor(assembly)

## This is used to store the frames
## for creating a movie
w2i = vtk.vtkWindowToImageFilter()
w2i.SetInput(renWin)
w2i.Update()

## The TIFF writer
writer = vtk.vtkTIFFWriter()
writer.SetInputConnection(w2i.GetOutputPort())
writer.SetCompressionToJPEG()

## Initialize camera
cam = ren.GetActiveCamera()
cam.SetFocalPoint(0.,0.,Lmax/2)
cam.SetViewUp(0.,1.,0.);

## The flying camera is as follows.
## It rotates around the focal point but also the distance to the
## object varies sinusoidally between Rcammin and Rcammax
Rcammax = 6*Lmax
Rcammin = 2*Lmax
## The speed of rotation of the camera
wcam = 0.05
## The speed with which the radius varies
wRcam = 1.33*wcam


# Initial frame
frame = 0
## Total nbr of frames
nframes = 300
## Store frames as TIFF images in ./YUV directory
save_frames = 0

# Loop while: rotating the camera and modify
# node coordinates
while 1:
    t += Dt
    ## time.sleep(0.3)
    if frame%10==0:
        print "frame: ",frame

    ## This recomputes the clipping plane Otherwise, since the camera
    ## is rotating, some objects may disappear
    ren.ResetCameraClippingRange()
    change_coords_filter.Modified()

    ## Actual length of worm
    L = z2-z1
    ## Recompute which extreme is moving
    if L>Lmax:
        moving = 1
    if L<Lmin:
        moving = 2
    ## Update positions of extremes
    if moving == 1:
        z1 += speed*Dt
    else:
        z2 += speed*Dt
    ## Substract mean velocity
    ## We describe the world from a system moving with the
    ## mean velocity of the worm, i.e. mspeed = speed/2
    z1 -= mspeed*Dt
    z2 -= mspeed*Dt

    ## Update camera
    phi = wcam*t
    Rcamav =(Rcammin+Rcammax)/2.0
    DRcam = (Rcammin-Rcammax)/2.0
    Rcam = Rcamav + DRcam*sin(wRcam*t)
    Hcam = 0.25*Rcam
    cam.SetPosition(Rcam*cos(phi),Hcam,Rcam*sin(phi))
    renWin.Render()

    ## Save current frame
    if save_frames:
        w2i.Modified()
        writer.SetFileName("./YUV/frame." + str(frame) +".tiff")
        writer.Write()

    if nframes>0 && frame == nframes:
        break
    
    ## Update the position of the cones.
    ## They move with velocity [0,0, -mspeed]. 
    ## In order to keep the cones inside the landscape
    ## when a cone exits the landscape (i.e. z_cone<-Lscape)
    ## then it is moved to the same position but at the inlet (i.e.
    ## z_cone = +Lscape
    for k in range(ncones):
        cone = cones[k]
        cone.AddPosition(0,0,-mspeed*Dt)
        x,y,z = cone.GetPosition()
        if z<-Lscape:
            z = Lscape
            cone.SetPosition(x,y,z)
            
    ## Update frame counter
    frame += 1
            
